module Tomte::Protocol
  module Dialog

    class Client < Tomte::Protocol::Producer

      autoload :Request,  'tomte/protocol/dialog/client/request'

      DEFAULT_DIALOG_PUBLISH_OPTIONS = { :mandatory => true }.freeze
      DEFAULT_RETURN_QUEUE_OPTIONS = { :auto_delete => true }.freeze

      DEFAULT_REQUEST_TIMEOUT = nil


      attr_accessor :dialogs_pending

      # A Dialog is a kind of Producer, that for every request, expects an answer to be given from the consumer
      # thus it instantiates a temporary queue for such requests to be given and adds a return address to the outgoing
      # requests.
      #
      # @example Instantiating a new Dialog::Client
      #
      #   errback = proc {|channel, channel_close| puts 'initialization failed!'}
      #
      #   client = Dialog::Client.new(connection
      #     :exchange => 'tomte.my.exchange',
      #     :exchange_type => :direct,
      #     :enveloper => Tomte::Protocol::Enveloper.new(:serialization => :bson)
      #
      #   ).connect!(:errback => errback) do |producer|
      #     puts "producer initialized successfully!"
      #   end
      #
      #   client.call(data) do |response|
      #     puts response
      #   end
      #
      #
      #
      # @param [Tomte::Endpoint] connection         the Endpoint you want to open this Producer on
      #
      # @option options [Exchange] :exchange        Either a Tomte::Protocol::Exchange, or a String with an exchange name,
      #                                             The Producer will create/instantiate an Exchange with the given option
      #                                             and publish subsequent messages using it.
      #                                             If no :exchange is given, the Producer will use the default AMQP direct
      #                                             exchange
      #
      # @option options [Symbol] :exchange_type     (:direct) The type for the exchange, if name is given,
      #
      # @option options [Hash] :exchange_options       The options using when declaring/instantiate the exchange.
      #                                             the same options you would pass to instantiate a [Tomte::Protocol::Exchange]
      #
      # @option options [Hash] :publish_options        Default options used when publishing a message from this producer
      #                                             merged to the options given on the #publish method
      #
      # @option options [Boolean] :reliable         Reliable Dialogs will use the confirm-select extension to ensure
      #                                             published messages are reliably delivered to the broker
      #                                             It will also set default options to.... (TODO check!)
      #                                             On reliable publishers, the publish callback is called _after_ the ack
      #                                             instead of when the publish-ok is returned
      #
      # @option options [Tomte::Protocol::Enveloper] :enveloper An instance of a [Tomte::Protocol::Enveloper] that will be
      #                                                         used to serialize/decorate message going though this [Entity]
      #
      #
      # @raise [AMQP::Error]  Will be re-raised from the AMQP layer
      #

      def initialize(connection, options = {})
        super

        @return_queue_options = options[:return_queue_options] || DEFAULT_RETURN_QUEUE_OPTIONS
        @return_queue_name = options[:return_queue_name] || "tomte.dlg.ret.#{random}"

        # override the default, a dialog may want a different default
        @publish_options = options[:publish_options] || DEFAULT_DIALOG_PUBLISH_OPTIONS

        @request_timeout = options[:request_timeout] || DEFAULT_REQUEST_TIMEOUT
      end


      #
      # Makes the Dialog to create the channel, and triggers any :after_channel_ready callbacks in descendants
      #
      # @option options [Proc] errback              Callback to call when the instantiation/creation is not successful
      #                                             will be called with |channel, channel_close|
      #
      # @yield [self] Yields successfully when channel is created, and the exchange successfully declared
      # @yieldparam [Entity] self This Entity that is successfully declared and is ready to be used.
      #
      # @api public
      #

      def connect!(options = {}, &block)
        super
        reset_dialogs_pending
        self
      end


      # Publishes message and waits for the answer.
      #
      # @example sending a request through a dialog
      #
      #   client.call(data) do |response|
      #     puts response
      #   end
      #
      #
      # @example sending a request with reliability
      #
      #   errback = proc { |error| puts "an error occurred during the request #{error.to_s}" }
      #   on_ack  = proc { puts "the request has been safely acked by the broker, but we haven't got an answer, yet'" }
      #   updates = proc { |u| puts "we got an update, server is working on our request!"}
      #
      #   reliable_client.call(request,
      #     :updates => updates,
      #     :errback => errback,
      #     :on_ack => on_ack,
      #     :mandatory => true
      #   ) do |response|
      #     # response from the server
      #     puts response
      #   end
      #
      # @param  [Object] payload  Message payload (content). This Object is given to the Tomte::Protocol::Enveloper for
      #                           serialization, before being sent to the actual exchange for publishing
      #
      # @option options [String] :routing_key (nil)  Specifies message routing key. Routing key determines
      #                                      what queues messages are delivered to (exact routing algorithms vary
      #                                      between exchange types).
      #
      # @option options [Boolean] :mandatory (false) This flag tells the server how to react if the message cannot be
      #                                      routed to a queue. If message is mandatory, the server will return
      #                                      unroutable message back to the client with basic.return AMQP method.
      #                                      If message is not mandatory, the server silently drops the message.
      #
      # @option options [Boolean] :immediate (false) This flag tells the server how to react if the message cannot be
      #                                      routed to a queue consumer immediately.  If this flag is set, the
      #                                      server will return an undeliverable message with a Return method.
      #                                      If this flag is zero, the server will queue the message, but with
      #                                      no guarantee that it will ever be consumed.
      #                                      NOTE THAT THIS IS NO LONGER SUPPORTED with RabbitMQ > 3.0
      #
      # @option options [Boolean] :persistent (false) When true, this message will be persisted to disk and remain in the queue until
      #                                       it is consumed. When false, the message is only kept in a transient store
      #                                       and will lost in case of server restart.
      #                                       When performance and latency are more important than durability, set :persistent => false.
      #                                       If durability is more important, set :persistent => true.
      #
      # @option options [String] :content_type (application/octet-stream) Content-type of message payload.
      #
      # @option options [Integer] :ack_timeout How many seconds to wait for an acknowledgement before considering the request unacked
      #                                        by the broker. This is about the reliable producers.
      #
      # @option options [Integer] :timeout    Wait this many seconds, then return the request with an error. If the server supports
      #                                       expiration, the expiration header is also set to this value (in seconds). This
      #                                       causes the message to be dequeued if it is not delivered before the client
      #                                       has timed out.
      #
      # @option options [Proc] :errback       Callback to call when the publish is not successful. Works in reliable mode.
      #                                       will be called with |error_message|, the amqp message, that caused the error
      #                                       possibly a basic.return or a channel.close, nil for a timeout
      # @option options [Proc] :on_ack        On reliable dialogs, it is called when the request has been acked by the broker.
      #
      # @option options [Proc] ;updates       A callback that gets called for request updates
      #
      # @return [Dialog::Client] self
      #
      # @api public
      #
      # @see Producer#publish
      #

      def call(object, options = {}, &block)
        check_ready

        request = Request.new(self, object, options, &block)
        timeout = options[:timeout] || options[:request_timeout] || @request_timeout
        request.updates = options.fetch(:updates, nil)

        if timeout
          request.timeout timeout
          options[:expiration] = timeout*1000
        end
        dialogs_pending[request.message_id] = request

        tomte_debug do
          log.debug "+ [#{@channel.id}] sending request. Message-ID: '#{request.message_id}', Timeout #{timeout}."
          log.debug options.ai.subindent
        end

        publish(request.payload,
                options.merge(:reply_to => request.replyable? ? @return_queue.name : nil,
                             :message_id => request.message_id, :headers => request.headers,
                             :type => 'tomte/dlg.request').merge(@publish_options),
                             &options[:on_ack])

      end


      private
      #######

      #
      # sets up this entity callback chain
      #
      def append_entity_callbacks

        append_callback(:after_channel_ready, &method(:setup_return_queue))
        append_callback(:after_return_queue_ready, &method(:subscribe_return_consumer))
        append_callback(:after_return_queue_ready, &method(:setup_exchange))

        append_callback(:after_exchange_ready) { entity_ready! }
      end

      #
      # reset the pending state
      #
      def reset_dialogs_pending
        @dialogs_pending = Hash.new
      end

      #
      # dialog message returned with response
      #
      def on_response(metadata, payload)
        if(r = @dialogs_pending.delete(metadata.correlation_id))

          r.returned(metadata, payload)
        end
      end

      #
      # dialog server sent an update
      #
      def on_update(metadata, payload)
        if(r = @dialogs_pending.fetch(metadata.correlation_id, nil))

          r.received_update(metadata, payload)
        end
      end

      #
      # the message failed even before publishing (likely on enveloping)
      #
      def on_publish_error(payload, options, error)
        if(r = @dialogs_pending.delete(options[:message_id]))

          r.returned_with_publish_error(error)
        end
      end

      #
      # we got an answer we can correlate, but we can't read it
      #
      def on_format_exception(metadata, payload, exception)
        if(r = @dialogs_pending.delete(metadata.correlation_id))

          r.returned_with_format_exception(metadata, payload, exception)
        end
      end

      #
      # an error happened on the channel
      #
      def on_error(ch, channel_close)
        return_all_pending_requests(channel_close)
        reset_dialogs_pending
        super
      end

      #
      # a request has returned with a failure
      #
      def on_return(basic_return, metadata, payload)
        super
        if(r = @dialogs_pending.delete(metadata.properties[:message_id]))
          r.returned_with_failure(metadata, payload, basic_return)
        end
      end

      #
      # channel has been restored, if we have pending request, fail them all
      #
      def return_all_pending_requests(error)
        @dialogs_pending.keys.key do |k|
          if(r = @dialogs_pending.delete(k))
            r.returned_with_failure(metadata, payload, error)
          end
        end
      end

      #
      # set-up a queue for returning responses
      #
      def setup_return_queue

         @return_queue = @channel.queue(@return_queue_name,
                                        @return_queue_options.dup) do |q, declare_ok|

           log.debug("+ [#{@channel.id}] set up return queue '#{@return_queue.name}', options: " +
                      @return_queue_options.inspect )
           exec_callback(:after_return_queue_ready)
         end

      end

      #
      # and subscribes on it
      #
      def subscribe_return_consumer

        log.debug("+ [#{@channel.id}] subscribing on '#{@return_queue.name}'")

        @return_queue.subscribe do |metadata, payload|

          log.debug("+ [#{@channel.id}] received response. Message-ID: '#{metadata.message_id}' ")

          tomte_debug do
            log.debug metadata.attributes.ai.subindent
            log.debug payload.ai.subindent
          end

          begin
            data = envelope_wire_to_local(payload, metadata)

            case data.metadata.type
              when nil, 'tomte/dlg.response', 'tomte/dlc.exception'
                on_response(data.metadata, data.payload)
              when 'tomte/dlg.update'
                on_update(data.metadata, data.payload)
            end

          rescue Tomte::Protocol::Enveloper::Error,
                 Tomte::Protocol::Serializer::Error => e

            log.warn "+ [#{@channel.id}] Format Exception"
            log.exception(e, :warn)

            on_format_exception(metadata, payload, e)
          end
        end
      end
    end
  end
end