module Tomte::Protocol
  module Dialog

    class Server < Tomte::Protocol::Consumer

      autoload :Request,  'tomte/protocol/dialog/server/request'
      autoload :Response, 'tomte/protocol/dialog/server/response'
      autoload :Update,   'tomte/protocol/dialog/server/update'

      RETURN_EXCHANGE      = AMQ::Protocol::EMPTY_STRING
      RETURN_EXCHANGE_TYPE = :direct

      DEFAULT_RETURN_PUBLISH_OPTIONS = { :mandatory => true }

      #
      # @example Instantiating a new Dialog::Server
      #
      #  errback = proc {|channel, channel_close| puts 'initialization failed!'}
      #
      #   server = Server.new(connection,
      #     :queue => 'tomte.my.queue',
      #     :exchange_type => :direct,
      #     :enveloper => Tomte::Protocol::Enveloper.new(:serialization => :bson)
      #
      #   )
      #   server.bind('tomte.my.exchange', :key => "my.key")
      #   server.on_call do |metadata, message|
      #
      #     # what you return in the block is returned
      #     # to the sender
      #     return :pong
      #   end
      #
      #   server.connect!(:errback => errback) do |server|
      #     puts "server initialized successfully!"
      #     consumer.consume
      #   end
      #
      # @param [Tomte::Endpoint] connection         the Endpoint you want to open this Producer on
      #
      # @option options [Queue] :queue               Either a Tomte::Protocol::Queue, or a String with a queue name,
      #                                             The Consumer will create/instantiate a Queue with the given option
      #
      # @option options [Hash] :queue_options       The options using when declaring/instantiate the queue.
      #                                             the same options you would pass to instantiate a [Tomte::Protocol::Queue]
      #
      # @option options [Hash] :consumer_options    Default options used when creating the AMQP consumer
      #
      # @option options [Boolean] :exclusive        Request exclusive consumer access, meaning only this consumer can access the queue.
      #                                             This is useful when you want a long-lived shared queue to be temporarily accessible by just
      #                                             one application (or thread, or process). If application exclusive consumer is part of crashes
      #                                             or loses network connection to the broker, channel is closed and exclusive consumer is thus cancelled.
      #
      #
      # @option options [Boolean] :ack              If this field is set to false the server does not expect acknowledgments
      #                                             for messages.  That is, when a message is delivered to the client
      #                                             the server automatically and silently acknowledges it on behalf
      #                                             of the client.  This functionality increases performance but at
      #                                             the cost of reliability.  Messages can get lost if a client dies
      #                                             before it can deliver them to the application.
      #
      # @option options [Boolean] :ack_requests     Automatically ack every request when the on_call block ends successfully
      #                                             if set to false, it's the block responsibility to explicitly ack the request
      #
      # @options options [Proc] :on_call            The callback to invoke for each request, same as #on_call
      #
      #
      # @option options [Boolean] :no_local         Request the consumer to be no_local
      #
      #
      #
      # @option options [Tomte::Protocol::Enveloper] :enveloper An instance of a [Tomte::Protocol::Enveloper] that will be
      #                                                         used to serialize/decorate message going though this [Entity]
      #
      #
      # @raise [AMQP::Error]  Will be re-raised from the AMQP layer
      #
      #
      def initialize(connection, options = {})

        super

        # override the queue name if it's the random queue
        @queue_name = "tomte.dlg.srv.#{random}" if(@queue_name == AMQ::Protocol::EMPTY_STRING)
        @return_publish_options = options[:return_publish_options] || DEFAULT_RETURN_PUBLISH_OPTIONS
        @ack_requests = options[:ack_requests] || @ack

        on_delivery &method(:call)

        if options[:on_call] && options[:on_call].respond_to?(:call)
          on_call(&options[:on_call])
        else
          on_call { log.warn "+ [#{@channel.id}] received request without a call handler."; nil }
        end
      end

      # Registers the callback that is called for every received message
      #
      # @yield [metadata, payload] Yielded for every received message
      # @yieldparam [metadata] AMQP metadata for the received message
      # @yieldparam [Object]  the received object, can be a  Tomte::Document, or Tomte:Message, or a ruby
      #                       object, according to how the enveloper was set
      # @yield [request]      Yielded for every received message
      # @yieldparam [request] A Request object containing both metadata and payload
      #
      # @return [Server] self
      # @api public
      #

      def on_call(&proc)
        @on_call = proc
        self
      end


      def call(metadata, payload)

        request = Request.new(metadata, payload)
        request.response_callback = method(:process_result)
        request.update_callback = method(:send_update)

        tomte_debug do
          log.debug "+ [#{@channel.id}] received a dialog request."
          log.debug metadata.attributes.ai.subindent
          log.debug payload.ai.subindent
        end

        catch(:_async) do
          response = @on_call.call(@on_call.arity == 1 ? request : [metadata, payload])
          process_result(request, response)
        end

      rescue Exception => exception

        log.exception exception, :warn
        process_result(request, exception)
      end

      #
      # like EM#defer, but with handling of exceptions
      #
      def self.protected_defer(job, callback = nil, errback = nil, &block)

        deferred_callback = proc do |r|
          case r
            when Exception
              errback.call(r) if errback
            else
              callback.call(r) if callback
          end
        end

        deferred_job = proc do
          begin
            (block || job).call
          rescue Exception => e
            e
          end
        end

        EM.defer(deferred_job, deferred_callback)
      end

      def protected_defer(job, callback = nil, errback = nil, &block)
        log.debug "+ [#{@channel.id}] request is going to be served in a deferred task."
        self.class.protected_defer(job, callback, errback, &block)
      end

      protected
      #########

      #
      # allows subclasses to perform postprocessing on response before sending them back
      #
      def postprocess_request(request); end


      private
      #######

      #
      # send an update before the final result
      #
      def send_update(request, update, options)
        additional_headers = options.delete(:additional_headers)

        send_response request, Update.new(update, options, additional_headers), true
      end

      #
      # does the processing after the server called back
      #
      def process_result(request, res)

        tomte_debug do
          log.debug "+ [#{@channel.id}] processing result."
          log.debug res.ai.subindent
        end

        # an async processor may catch an exception and return it as a message
        if res.kind_of? Exception
          request.response.additional_headers = build_headers_for(res)
        else
          request.response.payload ||= res
        end

        postprocess_request(request)

        if request.reject
          request.metadata.reject(request.reject)
        else
          send_response(request, request.response) if request.response.valid?
          request.metadata.ack unless (request.nak || !@ack_requests)
        end
      end


      #
      #
      #
      def send_response(request, response, update = false)

        if(sender = request.metadata.reply_to)

          options = response.options.merge(
              :headers => response.additional_headers
          )

          begin
            data = envelope_local_to_wire(response.payload, options)

          rescue Tomte::Protocol::Enveloper::Error,
                 Tomte::Protocol::Serializer::Error => e

            log.error "+ [#{@channel.id}] Enveloper Exception."
            log.exception e, :warn

            data = Tomte::Protocol::Enveloper::WireResponse.new('', {
              :headers => response.additional_headers = build_headers_for(e),
              :type => 'tomte/dlg.exception'
            })
          end

          data.metadata[:routing_key] = sender
          data.metadata[:correlation_id] = request.metadata.message_id
          data.metadata[:message_id] ||= "r.#{request.metadata.message_id}"
          data.metadata[:type] = update ? 'tomte/dlg.update' : 'tomte/dlg.response'

          log.debug "+ [#{@channel.id}] sending response. Message-ID: '#{data.metadata[:message_id]}' to #{sender}."
          tomte_debug { log.debug data.ai.subindent }

          @return_exchange.publish(data.body, @return_publish_options.merge(data.metadata))
        end
      end

      #
      # sets up this entity callback chain
      #
      def append_entity_callbacks

        append_callback(:after_channel_ready, &method(:setup_queue))
#        append_callback(:after_queue_ready, &method(:setup_consumer))
        append_callback(:after_queue_ready, &method(:setup_return_exchange))
#        append_callback(:after_consumer_ready ) { entity_ready! }
        append_callback(:after_queue_ready ) { entity_ready! }

      end

      #
      # gets the default exchange for returning the response
      #
      def setup_return_exchange

        @return_exchange = @channel.exchange(RETURN_EXCHANGE,
                                             RETURN_EXCHANGE_TYPE)

      end

      def build_headers_for(exception)

        {'tomte-dialog-exception' => {
           :type => exception.class.name,
           :remote_message => exception.message,
           :caption => exception.to_s
          }
        }
      end
    end
  end
end