
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol
  class Consumer < Entity

    DEFAULT_QUEUE_OPTIONS = { }.freeze
    DEFAULT_EXCHANGE_OPTIONS = { }.freeze
    DEFAULT_BIND_OPTIONS = { }.freeze

    DEFAULT_CONSUMER_OPTIONS = { }.freeze

    #
    # @example Instantiating a new Consumer
    #
    #  errback = proc {|channel, channel_close| puts 'initialization failed!'}
    #
    #   consumer = Consumer.new(connection,
    #     :queue => 'tomte.my.queue',
    #     :exchange_type => :direct,
    #     :enveloper => Tomte::Protocol::Enveloper.new(:serialization => :bson)
    #
    #   )
    #   consumer.bind('tomte.my.exchange', :key => "my.key")
    #   consumer.on_delivery do |metadata, message|
    #     puts "received a message!"
    #   end
    #
    #   consumer.connect!(:errback => errback) do |consumer|
    #     puts "producer initialized successfully!"
    #     consumer.consume
    #   end
    #
    # @param [Tomte::Endpoint] connection         the Endpoint you want to open this Producer on
    #
    # @option options [Queue] :queue               Either a Tomte::Protocol::Queue, or a String with a queue name,
    #                                             The Consumer will create/instantiate a Queue with the given option
    #
    # @option options [Hash] :queue_options       The options using when declaring/instantiate the queue.
    #                                             the same options you would pass to instantiate a [Tomte::Protocol::Queue]
    #
    # @option options [Hash] :consumer_options    Default options used when creating the AMQP consumer
    #
    # @option options [Boolean] :exclusive        Request exclusive consumer access, meaning only this consumer can access the queue.
    #                                             This is useful when you want a long-lived shared queue to be temporarily accessible by just
    #                                             one application (or thread, or process). If application exclusive consumer is part of crashes
    #                                             or loses network connection to the broker, channel is closed and exclusive consumer is thus cancelled.
    #
    #
    # @option options [Boolean] :ack              If this field is set to false the server does not expect acknowledgments
    #                                             for messages.  That is, when a message is delivered to the client
    #                                             the server automatically and silently acknowledges it on behalf
    #                                             of the client.  This functionality increases performance but at
    #                                             the cost of reliability.  Messages can get lost if a client dies
    #                                             before it can deliver them to the application.
    #
    #
    # @option options [Boolean] :no_local         Request the consumer to be no_local
    #
    #
    #
    # @option options [Tomte::Protocol::Enveloper] :enveloper An instance of a [Tomte::Protocol::Enveloper] that will be
    #                                                         used to serialize/decorate message going though this [Entity]
    #
    #
    # @raise [AMQP::Error]  Will be re-raised from the AMQP layer
    #
    #

    def initialize(connection, options = {})
      super

      options[:queue].is_a?(Queue) ?
        @queue = options[:queue] :
        (@queue_name = options[:queue] || AMQ::Protocol::EMPTY_STRING)

      @queue_options = options.fetch(:queue_options, DEFAULT_QUEUE_OPTIONS)

      @exclusive     = options.fetch(:exclusive,     false)
      @ack           = options.fetch(:ack,           true)
      @no_local      = options.fetch(:no_local,      false)

      @consumer_options = options.fetch(:default_consumer_options, DEFAULT_CONSUMER_OPTIONS)

      if options[:on_delivery]
        on_delivery(&options[:on_delivery])
      else
        @on_delivery = proc { |message| log.info "+ [#{@channel.id}] received a message without a message handler." }
      end
      @state.consuming = false

      append_entity_callbacks
    end

    # @option options [Proc] :errback             Callback to call when the instantiation/creation is not successful
    #                                             will be called with |channel, channel_close|
    #
    # @yield [self] Yields successfully when channel is created, and the exchange successfully declared
    # @yieldparam [Producer] self This producer that is successfully declared and is ready to be used.
    #
    # @api public
    #

    def connect!(options = {}, &block)
      super
    end

    def pop(options = {}, &block)
      check_ready

      @queue.pop(options) do |metadata, payload|

        ## TODO handle envelope errors

        data = envelope_wire_to_local(payload, metadata)
        case block.arity
        when 1 then
          block.call(data.payload)
        when 2 then
          block.call(data.metadata, data.payload)
        end

      end
    end


    # Begin consuming messages from the queue
    # @return [Consumer] self
    # @api public
    #

    def consume(options = {}, &block)
      check_ready

      raise(Error, "Consumer is already subscribed") if @state.consuming

      setup_consumer(options)

      @consumer.consume(options[:nowait]) do
        @state.consuming = true
        log.debug "+ [#{@channel.id}] consuming on queue '#{@queue.name}'"
        block.call(self) if block
      end
      self
    end

    #
    # connect and consume
    #
    def ready!(options = {}, &block)

      connect!(options) do
        consume(&block)
      end
    end


    # Cancels consuming messages from the queue
    # @return [Consumer] self
    # @api public
    #

    def cancel(nowait = false, &block)
      check_ready

      @consumer.cancel(nowait) do

        @state.consuming = false
        log.debug "+ [#{@channel.id}] stopped consuming on queue '#{@queue.name}'"
        block.call if block
      end
      self
    end


    #
    # Declares a new exchange in the context of this consumer.
    #
    # @see Exchange#initialize
    # @api public
    #

    def declare_exchange(name, type = :direct, options = {}, &block)
      check_ready

      exchange_options = DEFAULT_EXCHANGE_OPTIONS.merge(options)

      log.debug "+ [#{@channel.id}] declaring exchange '#{name}', type: #{type} options: " +
                exchange_options.inspect

      @channel.exchange(name, type, exchange_options, &block)
    end


    # This method binds the consumer's queue to an exchange. Until the queue is
    # bound it will not receive any messages.
    #
    # A valid exchange name (or reference) must be passed as the first
    # parameter.
    # @example Binding the consumer to exchange using an Exchange instance
    #
    #  ch       = AMQP::Channel.new(connection)
    #  exchange = ch.direct('backlog.events')
    #  queue    = ch.queue('', :exclusive => true)
    #  queue.bind(exchange)
    #
    #
    # @example Binding a queue to exchange using exchange name
    #
    #  ch       = AMQP::Channel.new(connection)
    #  queue    = ch.queue('', :exclusive => true)
    #  queue.bind('backlog.events')
    #
    #
    # Note that if your producer application knows consumer queue name and wants to deliver
    # a message there, direct exchange may be sufficient (in other words, if your code declares an exchange with
    # the same name as a queue and binds it to that queue, consider using the default exchange and routing key on publishing).
    #
    # @param [Exchange] Exchange to bind to. May also be a string or any object that responds to #name.
    #
    # @option options [String] :routing_key   Specifies the routing key for the binding. The routing key is
    #                                         used for routing messages depending on the exchange configuration.
    #                                         Not all exchanges use a routing key! Refer to the specific
    #                                         exchange documentation.  If the routing key is empty and the queue
    #                                         name is empty, the routing key will be the current queue for the
    #                                         channel, which is the last declared queue.
    #
    # @option options [Hash] :arguments (nil)  A hash of optional arguments with the declaration. Headers exchange type uses these metadata
    #                                          attributes for routing matching.
    #                                          In addition, brokers may implement AMQP extensions using x-prefixed declaration arguments.
    #
    # @option options [Boolean] :nowait (true)  If set, the server will not respond to the method. The client should
    #                                           not wait for a reply method.  If the server could not complete the
    #                                           method it will raise a channel or connection exception.
    # @return [Queue] Self
    #
    #
    # @yield [] Since queue.bind-ok carries no attributes, no parameters are yielded to the block.
    #
    # @api public
    # @see Queue#bind
    #

    def bind(exchange, options = {}, &block)
      check_ready

      log.debug "+ [#{@channel.id}] bind on '#{exchange.respond_to?(:name) ? exchange.name : exchange }', options: " +
                options.inspect

      @queue.bind(exchange, DEFAULT_BIND_OPTIONS.merge(options), &block)
    end


    # Remove the binding between the consumer and exchange. The queue will
    # not receive any more messages until it is bound to another
    # exchange.
    #
    # Due to the asynchronous nature of the protocol, it is possible for
    # "in flight" messages to be received after this call completes.
    # Those messages will be serviced by the last block used in a
    # {Queue#subscribe} or {Queue#pop} call.
    #
    # @param [Exchange] Exchange to unbind from.
    #
    # @option opts [Boolean] :nowait (true)  If set, the server will not respond to the method. The client should
    #                                       not wait for a reply method.  If the server could not complete the
    #                                       method it will raise a channel or connection exception.
    #
    #
    # @yield [] Since queue.unbind-ok carries no attributes, no parameters are yielded to the block.
    #
    # @api public
    # @see Queue#unbind
    #

    def unbind(exchange, options = {}, &block)

      log.debug "+ [#{@channel.id}] un-bind on '#{exchange.respond_to?(:name) ? exchange.name : exchange }', options: " +
                 options.inspect

      @queue.unbind(exchange, options, &block)
    end

    # Registers the callback that is called for every received message
    #
    # @yield [metadata, payload] Yielded for every received message
    # @yieldparam [metadata] AMQP metadata for the received message
    # @yieldparam [Object]  the received object, can be a  Tomte::Document, or Tomte:Message, or a ruby
    #                       object, according to how the enveloper was set
    #
    # @return [Consumer] self
    # @api public
    #

    def on_delivery(&block)

      @on_delivery = Proc.new { | headers, payload, consumer_tag, delivery_tag, redelivered, exchange, routing_key |

        begin
          data = envelope_wire_to_local(payload, headers,
                                        :consumer_tag => consumer_tag,
                                        :delivery_tag => delivery_tag,
                                        :redelivered => redelivered,
                                        :exchange => exchange,
                                        :routing_key => routing_key)

          log.debug "+ [#{@channel.id}] received a message: Message-Id: #{headers.message_id}."
          log.debug "-     from exchange: '#{exchange}', key: '#{routing_key}'"

          exec_callback(:on_delivery, data.metadata, data.payload)

          case block.arity
          when 1 then
            block.call(data.payload)
          when 2 then
            block.call(data.metadata, data.payload)
          else
            block.call(data.metadata, data.payload, consumer_tag,
                     delivery_tag, redelivered, exchange, routing_key)
          end

        # something went wrong with de-enveloping
        rescue Tomte::Protocol::Enveloper::Error,
               Tomte::Protocol::Serializer::Error => e

          log.warn "+ [#{@channel.id}] Enveloper Exception"
          log.exception(e, :warn)
        end

      }

      self
    end

    def delete_queue(options = {}, &block)

      check_ready

      @queue.delete do |delete_ok|

        log.debug "+ [#{@channel.id}] deleted queue."
        block.call delete_ok if block
      end
    end

    private
    #######

    #
    # sets up this entity callback chain
    #
    def append_entity_callbacks

      append_callback(:after_channel_ready, &method(:setup_queue))
#      append_callback(:after_queue_ready, &method(:setup_consumer))
#      append_callback(:after_consumer_ready ) { entity_ready! }
      append_callback(:after_queue_ready ) { entity_ready! }
    end



    #
    # sets up this consumer's queue
    #
    def setup_queue

      @queue = @channel.queue(@queue_name,
                              @queue_options.dup) do |q, declare_ok|

        log.debug "+ [#{@channel.id}] set up queue '#{q.name}', options:" + @queue_options.inspect

        exec_callback(:after_queue_ready)
      end

    end

    #
    # sets up the underlying AMQP consumer
    #
    def setup_consumer(options = {})

      @consumer = Tomte::Protocol::Async::Consumer.new(@channel,
                                                       @queue,
                                                       Tomte::Protocol::Async::Consumer.tag_generator.generate_for(@queue),
                                                       @exclusive,
                                                       !@ack,
                                                       @consumer_options.merge(options),
                                                       @no_local, &options[:confirm])
      @consumer.on_delivery(&(@on_delivery))

      log.debug "+ [#{@channel.id}] set up consumer on queue '#{@queue.name}: {ack: #{@ack}, exclusive: #{@exclusive}}'."
    #  exec_callback(:after_consumer_ready)
    end

  end
end
