
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol

  autoload :Pending,      'tomte/protocol/entities/pending'

  class Entity

    class Error < StandardError; end
    class NotReadyError < Error; end

    include Tomte::Protocol::Async::Callbacks
    include Tomte::Protocol::Enveloper::Hooks

    DEFAULT_REPLY_TEXT = "Goodbye."

    REQUIRED_SERVER_CAPABILITIES = ['publisher_confirms', 'basic.nack', 'consumer_cancel_notify'].freeze

    SUGGESTED_SERVER_PRODUCT = 'RabbitMQ'
    SUGGESTED_SERVER_VERSION = '3.0.0'

    # the async_endpoint this entity is bound to
    attr_reader :async_endpoint

    # the channel this entity is working with
    attr_reader :channel

    # an instance of the logger this entity uses to log events
    attr_reader :log

    # the [Tomte::Enveloper] that serializes message going through this Entity
    attr_accessor :enveloper

    # Options for this Entity
    attr_reader :opts

    # State of this Entity
    attr_reader :state

    # Broker behaviour flags
    attr_reader :broker_behaviours


    #
    #
    #
    # @param [Tomte::Endpoint] endpoint           the Endpoint you want to open this Entity on
    #
    # @param [Hash] options                       an hash of options
    #
    # @option options [Tomte::Logger] :logger     (Tomte.logger) A Logger instance used by this entity to log messages
    #                                             will take the default Tomte.logger if none is given
    #
    # @option options [Tomte::Protocol::Enveloper] :enveloper An instance of a [Tomte::Protocol::Enveloper] that will be
    #                                                         used to serialize/decorate message going though this [Entity]
    #
    # @option options [Integer] :prefetch Set the channel prefetch to this size.
    #

    def initialize(endpoint, options = {}, &block)

      initialize_callbacks

      raise ArgumentError, "using an instance of AMQP::Connection is not allowed anymore" unless
          endpoint.respond_to?(:async) && endpoint.respond_to?(:client)
      @async_endpoint = endpoint.async

      @log = options.delete(:log) || Tomte.logger
      @enveloper = options[:enveloper]
      @prefetch = options[:prefetch]
      @opts = OpenStruct.new
      @state = OpenStruct.new
      @broker_behaviours = []

      state.ready = false
    end

    #
    # Makes the Entity to create the channel, and triggers any :after_channel_ready callbacks in descendants
    #
    # @option options [Proc] errback              Callback to call when the instantiation/creation is not successful
    #                                             will be called with |channel, channel_close|
    #
    # @yield [self] Yields successfully when channel is created, and the exchange successfully declared
    # @yieldparam [Entity] self This Entity that is successfully declared and is ready to be used.
    #
    # @api public
    #

    def connect!(options = {}, &block)

      raise(Error, "Already Connected!") if connected?

      append_callback(:after_channel_ready) { prefetch @prefetch } unless  @prefetch.nil?
      append_callback(:entity_ready, &block) if block_given?

      if options[:errback] && options[:errback].respond_to?(:call)
        append_callback(:failure_entity_initialization, &options[:errback])
      end

      if !async_endpoint.connected? && async_endpoint.respond_to?(:connect!)
        async_endpoint.connect! { connect_channel(async_endpoint)}
      else
        connect_channel(async_endpoint)
      end
      self
    end


    # Closes the Channel, then the Entity
    #
    # @api public
    #

    def close(reply_code = 200, reply_text = DEFAULT_REPLY_TEXT, &block)

      if connected?
        EM.next_tick do
          @channel.close(reply_code, reply_text) do |ch|

            log.debug "- [#{@channel.id}] entity closed successfully."

            block.call(self, ch) if block_given?
          end if @channel && @channel.open?
        end
      else
        block.call(self, nil) if block_given?
      end
      self
    end


    def connected?
      state.ready
    end

    def prefetch(size, global = false, &block)
      raise(NotReadyError, "#{self.class.name} channel is not declared.") if @channel.nil?
      channel.prefetch(size, global) do |channel|
        log.debug "- [#{@channel.id}] entity prefetch set to: #{size}."
        block.call self if block_given?
      end
    end

    private

    def check_ready
      raise(NotReadyError, "#{self.class.name} is not ready.") unless state.ready
    end

    def entity_ready!
      state.ready = true
      clear_callbacks(:failure_entity_initialization)
      exec_callback(:entity_ready, self)
    end

    def connect_channel(endpoint)
      @channel = endpoint.new_channel(:auto_recovery => true) do |ch, open_ok|
        check_server channel.connection

        log.debug("+ [#{@channel.id}] new #{self.class.name} requested channel #{channel.id}. Autorecovery: #{ch.auto_recovering?}")
        exec_callback_once(:after_channel_ready)
      end

      @channel.on_error &method(:on_error)

      if opts.confirm_select
        @channel.confirm_select do |confirm_select|
          log.debug("+ [#{@channel.id}] enabled confirm_select on channel.")
        end

        @channel.on_nack  &method(:on_nack)
        @channel.on_ack   &method(:on_ack)
      end
    end

    ##
    # Default CBs
    #

    def on_ack(basic_ack)

      log.debug({:event => :ack, :message => basic_ack}.ai.subindent)
    end


    def on_nack(basic_nack)

      log.debug({:event => :nack, :message => basic_nack}.ai.subindent)
    end


    def on_error(ch, channel_close)

      log.warn({:event => :error, :channel => ch.id, :message => channel_close}.ai.subindent)

      if has_callback?(:failure_entity_initialization)
        # entity is *still* initializing
        exec_callback_once(:failure_entity_initialization, ch, channel_close)
      else
        exec_callback(:channel_error, ch, channel_close, self)
        #ch.reuse
      end

    end

    def random
      Tomte.random_id
    end

    def check_server(connection)

      REQUIRED_SERVER_CAPABILITIES.each do |capability|
        if !connection.server_capabilities.include?(capability)
          log.fatal "+ [#{@channel.id}] AMQP Server does not support required capability '#{capability}'".important!
          return false
        end
      end

      if(!connection.broker || !connection.broker.properties ||
          connection.broker.properties['product'] != SUGGESTED_SERVER_PRODUCT)

        log.warn "+ [#{@channel.id}] AMQP Server does not look like #{SUGGESTED_SERVER_PRODUCT}. This has not been tested.".yellow
      elsif(connection.broker.properties['version'].nil? ||
          Gem::Version.new(connection.broker.properties['version']) <
          Gem::Version.new(SUGGESTED_SERVER_VERSION))

        broker_behaviours << :immediate_publish
        log.debug "+ [#{@channel.id}] Detected RabbitMQ version < #{SUGGESTED_SERVER_VERSION}.".yellow
      else
        log.debug "+ [#{@channel.id}] Detected RabbitMQ version #{SUGGESTED_SERVER_VERSION}, or greater.".yellow
        broker_behaviours << :message_expiration
      end
      true
    end

  end
end
