# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol
  module Pending
    extend ActiveSupport::Concern


    Event = Struct.new(:data, :message_id, :callback, :errback, :returned, :insertion_time)

    included do
      # Pending published messages
      attr_reader   :pending
    end

    private
    #######

    def reset_pending
      pending.last_acked = 0 if pending
    end

    def initialize_pending
      @pending = Struct.new(:by_id, :by_sequence, :by_expiration, :last_acked).new(
          {}, {}, Tomte::CoreStructures::PriorityQueue.new, 1)
    end

    def nack_by_timeout(time)

      #log.debug("- [#{@channel.id}] pending: #{pending.by_id.length}")

      while (!pending.by_expiration.empty? && pending.by_expiration.min_priority < time)
        nack_pending(pending.by_expiration.min_key)
      end
    end

    def pop_pending(sequence_number)
      event = pending.by_sequence.delete(sequence_number)
      pending.by_expiration.delete(sequence_number)
      pending.by_id.delete(event.message_id) if event
    end

    def push_pending(sequence_number, data, callback, errback, timeout = nil)
      pending.by_expiration.push(sequence_number, Time.now.to_i + timeout) if timeout
      pending.by_sequence[sequence_number] =
          pending.by_id[data.metadata[:message_id]] =
              Event.new(data, data.metadata[:message_id], callback, errback, nil, Time.now)
    end

    def nack_pending(delivery_tag, error = :timeout)

      log.debug "- [#{@channel.id}] pending message with tag: #{delivery_tag}, nacked by #{error.to_s}."
      return_pending(pop_pending(delivery_tag), error)
    end

    def ack_pending(delivery_tag)
      log.debug "- [#{@channel.id}] received acknowledge for message with delivery tag: #{delivery_tag}."

      state.last_acked_delivery_tag = delivery_tag
      return_pending pop_pending(delivery_tag)
    end

    def return_all_pending(error)
      log.debug "- [#{@channel.id}] returning all pending messages."
      pending.by_id.keys.each { |p| return_pending(pop_pending(p), error) }
    end

    def return_pending(event, error = nil)
      return unless event

      log.debug "- [#{@channel.id}] returning pending event #{event.message_id} after #{Time.now - event.insertion_time}s."

      if error == :timeout && event.errback
        event.errback.call(Tomte::Protocol::Timeout.new)
      elsif error && event.errback
        event.errback.call(
            Tomte::Protocol::TransportFailure.new("A Transport Failure Occurred",
                                                  :transport_exception => error))
      elsif event.returned && event.errback
        event.errback.call(
            Tomte::Protocol::TransportFailure.new("The message has been returned because of an exception",
                                                  :transport_exception => event.returned))
      else
        event.callback.call if event.callback
      end
    end

  end
end
