
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

# TODO
# - there a bug/strangeness in the AMQP library that makes channel.publisher_index not incremented
#   something fishey in the confirm extension
# - timeouts
#   publishes should have a timeout, after which the message is nack-ed



module Tomte::Protocol
  class Producer < Entity

    include Tomte::Protocol::Pending

    DEFAULT_PUBLISH_OPTIONS = { :default  => {},
                                :reliable => {:mandatory => true, :persistent => true }}.freeze

    DEFAULT_EXCHANGE_OPTIONS = { }.freeze

    DEFAULT_EXPIRATION  = nil
    DEFAULT_ACK_TIMEOUT = 30
    SWEEP_TIMER         = 5

    # The exchange this Producer works on
    attr_reader   :exchange

    #
    # @example Instantiating a new Producer
    #
    #  errback = proc {|channel, channel_close| puts 'initialization failed!'}
    #
    #   Producer.new(connection
    #     :exchange => 'tomte.my.exchange',
    #     :exchange_type => :direct,
    #     :enveloper => Tomte::Protocol::Enveloper.new(:serialization => :bson)
    #
    #   ).connect!(:errback => errback) do |producer|
    #     puts "producer initialized successfully!"
    #   end
    #
    # @param [Tomte::Endpoint] connection         the Endpoint you want to open this Producer on
    #
    # @option options [Exchange] :exchange        Either a Tomte::Protocol::Exchange, or a String with an exchange name,
    #                                             The Producer will create/instantiate an Exchange with the given option
    #                                             and publish subsequent messages using it.
    #                                             If no :exchange is given, the Producer will use the default AMQP direct
    #                                             exchange
    #
    # @option options [Symbol] :exchange_type     (:direct) The type for the exchange, if name is given,
    #
    # @option options [Hash] :exchange_options       The options using when declaring/instantiate the exchange.
    #                                             the same options you would pass to instantiate a [Tomte::Protocol::Exchange]
    #
    # @option options [Hash] :publish_options        Default options used when publishing a message from this producer
    #                                             merged to the options given on the #publish method
    #
    # @option options [Boolean] :reliable         Reliable publisher will use the confirm-select extension to ensure
    #                                             published messages are reliably delivered to the broker
    #                                             It will also set default options to.... (TODO check!)
    #                                             On reliable publishers, the publish callback is called _after_ the ack
    #                                             instead of when the publish-ok is returned
    #
    # @option options [Integer] :default_expiration   Default expiration time for messages published. In milliseconds.
    #
    # @option options [Tomte::Protocol::Enveloper] :enveloper An instance of a [Tomte::Protocol::Enveloper] that will be
    #                                                         used to serialize/decorate message going though this [Entity]
    #
    #
    # @raise [AMQP::Error]  Will be re-raised from the AMQP layer
    #

    def initialize(connection, options = {})
      super

      @opts.reliable = options[:reliable] || false

      if reliable?
        initialize_pending

        opts.confirm_select = true
      end

      options[:exchange].is_a?(Tomte::Protocol::Exchange) ?
        @exchange = options[:exchange] :
        @exchange_name = options[:exchange]

      @exchange_type =            options.fetch(:exchange_type,       :direct)
      @exchange_options =         options.fetch(:exchange_options,    DEFAULT_EXCHANGE_OPTIONS)
      @opts.default_ack_timeout = options.fetch(:default_ack_timeout, DEFAULT_ACK_TIMEOUT)
      @opts.default_expiration =  options.fetch(:default_expiration,  (reliable? ? DEFAULT_EXPIRATION : nil))

      @publish_options = options[:publish_options] || DEFAULT_PUBLISH_OPTIONS[(reliable? ? :reliable : :default )]

      unless (@exchange || @exchange_name)
        log.debug "+ using the default AMQP exchange"
        @exchange_name = AMQ::Protocol::EMPTY_STRING
        @exchange_type, @exchange_options = [:direct, {}]
      end

      append_entity_callbacks
    end


    # @option options [Proc] :errback             Callback to call when the instantiation/creation is not successful
    #                                             will be called with |channel, channel_close|
    #
    # @yield [self] Yields successfully when channel is created, and the exchange successfully declared
    # @yieldparam [Producer] self This producer that is successfully declared and is ready to be used.
    #
    # @api public
    #

    def connect!(options = {}, &block)
      @sweep_timer = Tomte::Timer.periodic(SWEEP_TIMER, &method(:sweep_timer)) if reliable?
      super
    end

    # Publishes message to the Producer. The message will be routed to queues by the exchange
    # and distributed to any active consumers. Routing logic is determined by exchange type and
    # configuration as well as message attributes (like :routing_key or message headers).
    #
    #
    #
    # h2. Data serialization
    #
    # Published data will be passed to the given Tomte::Enveloper instance before being actually published
    # to the exchange. If no enveloper is given, data is just published as-is. (just as if Exchange.publish has been called)
    #
    #
    #
    # h2. Reliability
    #
    # On reliable publishers, the publish callback is called _after_ the ack instead of when the publish-ok is returned
    #
    #
    # @param  [Object] payload  Message payload (content). This Object is given to the Tomte::Protocol::Enveloper for
    #                           serialization, before being sent to the actual exchange for publishing
    #
    # @option options [String] :routing_key (nil)  Specifies message routing key. Routing key determines
    #                                      what queues messages are delivered to (exact routing algorithms vary
    #                                      between exchange types).
    #
    # @option options [Boolean] :mandatory (false) This flag tells the server how to react if the message cannot be
    #                                      routed to a queue. If message is mandatory, the server will return
    #                                      unroutable message back to the client with basic.return AMQP method.
    #                                      If message is not mandatory, the server silently drops the message.
    #
    # @option options [Boolean] :immediate (false) This flag tells the server how to react if the message cannot be
    #                                      routed to a queue consumer immediately.  If this flag is set, the
    #                                      server will return an undeliverable message with a Return method.
    #                                      If this flag is zero, the server will queue the message, but with
    #                                      no guarantee that it will ever be consumed.
    #
    # @option options [Boolean] :persistent (false) When true, this message will be persisted to disk and remain in the queue until
    #                                       it is consumed. When false, the message is only kept in a transient store
    #                                       and will lost in case of server restart.
    #                                       When performance and latency are more important than durability, set :persistent => false.
    #                                       If durability is more important, set :persistent => true.
    #
    # @option options [String] :content_type (application/octet-stream) Content-type of message payload.
    #
    # @option options [Integer] :ack_timeout How many seconds to wait for an acknowledgement before considering the request unacked
    #                                        by the broker. This is about the reliable producers.
    #
    # @option options [Integer] :expiration In milliseconds. Sets the AMQP expiration header to this value, causing the message to
    #                                       be discarded if it has not been delivered to a client il the indicated time.
    #
    # @option options [Proc] :errback       Callback to call when the publish is not successful. Works in reliable mode.
    #                                       will be called with |exception|, a [Tomte::Protocol::Exception],
    #
    # @yield                                Callback to call when the publish is successful
    #                                       - in unreliable mode: means the message has been successfully staged for the broker
    #                                         the same as the Exchange#publish callback
    #                                       - in reliable mode: the message has been ack-ed positively, and no returns or errors
    #                                         have happened before the ack.
    #                                         Note that, returned messages are acked anyways, when the basic.return is confirmed.
    #                                         We consider any returned message to be a failure and the errback is called *when*
    #                                         the message is finally acked.
    #
    # @example Publishing without routing key
    #  producer = Producer.new(connection, :exchange => "tomte.exchange", :exchange_type => :fanout )
    #  # fanout exchanges deliver messages to bound queues unconditionally,
    #  # so routing key is unnecessary here
    #  producer.publish("some data")
    #
    # @example Publishing with a routing key
    #  Producer.new(connection, :exchange => "tomte.direct.exchange") do |producer|
    #    producer.publish("some data", :routing_key => "search.index.updates")
    #  end
    #
    # @return [Producer] self
    #
    # @api public
    #
    #

    def publish(payload, options = {}, &callback)

      # TODO WARNING
      # channel.publisher_index *SHOULD* be incremented by the channel, according by the *documentation*
      # and the code itself. Something is going wrong in the amq/client/async/extensions/rabbitmq/confirm.rb
      # see Extensions::RabbitMQ::Confirm::ExchangeMixin
      #
      # It should have no adverse effects, except when the bug is fixed.
      #
      # 2012-01-19 amq-client 0.9.1 fixes the increment thing but seems to offset the initial publisher_index
      # starting from 1 to 0, so we're sending. 0..1..2.. and rabbitmq is acking 1..2..3...
      # our sequence number became
      # sequence = (channel.publisher_index + 1)

      check_ready

      # rabbitmq > 3.0 does not support immediate mode publish anymore
      if (!broker_behaviours.include?(:immediate_publish) &&
        (immediate = options.delete(:immediate)))
          log.warn "+ [#{@channel.id}] Immediate publish is unsupported on this server."
      end

      errback = options.delete(:errback)
      timeout = options.delete(:ack_timeout) || @opts.default_ack_timeout
      options[:expiration] ||= @opts.default_expiration if @opts.default_expiration

      begin
        data = envelope_local_to_wire(payload, @publish_options.merge(options))

      rescue Tomte::Protocol::Enveloper::Error,
             Tomte::Protocol::Serializer::Error => e

        log.error "+ [#{@channel.id}] Enveloper Exception."
        log.exception(e, :warn)
        exception =  Tomte::Protocol::FormatException.new("Enveloper Exception", :format_exception => e)

        EM.next_tick do
          on_publish_error(payload, options, exception)
          errback.call exception if errback
        end

        return self
      end

      if reliable?
        sequence = (channel.publisher_index + 1)

        data.metadata[:message_id] ||= "tomte.rel.msg.#{random}"
        push_pending(sequence, data, callback, errback, timeout)

        log.debug "+ [#{@channel.id}] reliably publishing message. Message-ID: '#{data.metadata[:message_id]}', Tag: #{sequence}"
        @exchange.publish(data.body, data.metadata) { } ## { channel.publisher_index += 1 }
      else
        log.debug "+ [#{@channel.id}] publishing message. Message-ID: '#{data.metadata[:message_id]}'"
        @exchange.publish(data.body, data.metadata, &callback)
      end

      log.debug " -     on exchange '#{exchange.name}', wth routing key: '#{data.metadata[:routing_key]}'."
      self

    end

    # This method deletes an exchange. When an exchange is deleted all queue bindings on the exchange are deleted, too.
    # Further attempts to publish messages to a deleted exchange will result in a channel-level exception.
    # This effectively makes this Publisher unusable. It is intended to be used before close to make sure the exchange
    # that was created during initialization is deleted. (#close *won't* delete any exchange)
    #
    # @example Deleting an exchange
    #
    #  producer = Producer.new(connection, :exchange => "tomte.exchange")
    #  producer.delete_exchange { producer.close }
    #
    # @option options [Boolean] :nowait (false) If set, the server will not respond to the method. The client should
    #                                        not wait for a reply method.  If the server could not complete the
    #                                        method it will raise a channel or connection exception.
    #
    # @option options [Boolean] :if_unused (false) If set, the server will only delete the exchange if it has no queue
    #                                           bindings. If the exchange has queue bindings the server does not
    #                                           delete it but raises a channel exception instead.
    #
    # @return [NilClass] nil
    # @api public
    #

    def delete_exchange(options = {}, &block)

      check_ready

      @exchange.delete do |delete_ok|

        log.debug "+ [#{@channel.id}] deleted exchange."
        block.call delete_ok if block
      end
    end

    #
    # @return [Boolean] true if this Producer is reliable
    #

    def reliable?
      @opts.reliable
    end

    private
    #######

    #
    # sets up this entity callback chain
    #
    def append_entity_callbacks

      append_callback(:after_channel_ready, &method(:setup_exchange))
      append_callback(:after_exchange_ready) { entity_ready! }
    end


    #
    # ran every SWEEP_TIME seconds to check if there are any expired
    # messages still waiting for an ack
    #

    def sweep_timer
      nack_by_timeout(Time.now.to_i)
    end


    #
    # called for every basic.ack message
    #

    def on_ack(basic_ack)

      super
      if basic_ack.multiple
        ((pending.last_acked + 1) .. basic_ack.delivery_tag).each do |tag|
          ack_pending(tag)
        end
      else
        ack_pending(basic_ack.delivery_tag)
      end

    end

    #
    # an error that occurs while we're still publishing (ie: a serialization error)
    #

    def on_publish_error(payload, options, error)

      # do nothing?
    end


    #
    # called for any channel error
    #

    def on_error(ch, channel_close)
      super

      return_all_pending(channel_close) && reset_pending if reliable?
      @exchange = nil
    end


    #
    # called for any basic.return
    #

    def on_return(basic_return, metadata, payload)

      id = metadata.properties[:message_id]

      log.warn "- [#{@channel.id}] returned message '#{id}', #{basic_return.reply_code}: #{basic_return.reply_text}."
      log.debug basic_return.ai.subindent
      #log.debug metadata.ai.subindent

      if reliable?
        if(event = pending.by_id[id])
          event.returned = basic_return
        end
      end
    end


    #
    # sets things up
    #

    def setup_exchange
      if @exchange
        log.debug("+ [#{@channel.id}] using given exchange on '#{@exchange.name}'.")
        entity_ready!
      else
        @exchange = @channel.exchange(@exchange_name,
                                      @exchange_type,
                                      @exchange_options.dup ) do |exchange, declare_ok|

          log.debug("+ [#{@channel.id}] set up a new #{exchange.type} exchange '#{exchange.name}', options: " +
                          @exchange_options.inspect) if declare_ok

          exec_callback(:after_exchange_ready)
        end

      end

      @exchange.on_return &method(:on_return)
    end

  end
end