# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol
  module Enveloper

    class Generic

      include Utils

      attr_accessor :log, :trace
      attr_reader   :serialization

      def initialize(options = {})

        @log = options[:log] || Tomte.logger
        @trace = Tomte.config.protocol[:trace] || false
        @serialization = (options[:serialization] || Tomte.config.protocol[:serialization] || :bson).to_sym
        @message = options[:message] || false

        raise ArgumentError, "Unknown Serialization: '#{serialization}'" unless
            Serializer::SERIALIZERS.has_key?(serialization)

        log.debug "- Using the generic enveloper. Message output is #{@message}."
        log.warn  "= Message tracing is enabled, this will severely impact performance.".important! if trace?

      end

      #
      # converts the local representation of al object to the representation we intend to use on the wire
      #
      def local_to_wire(message, options = {})

        response = WireResponse.new

        case message
          when Tomte::Message

            response.payload = message.payload

            headers = (message.headers.to_h || {}).merge(options.delete(:headers) || {})
            response.metadata = message.metadata.to_wire.merge(options)
            response.metadata[:headers] = headers

          when Tomte::Document, Hash, nil
            response.metadata = options
            response.payload = message

          else
            response.metadata = options
            if message.respond_to?(:to_tomte) && (value = message.to_tomte).is_a?(Hash)
              value[:_type] ||= message.class.name
              response.payload = value
            elsif message.respond_to? :to_h
              response.payload = message.to_h
            else
              raise TypeError, "The enveloper doesn't know how to envelope a '#{message.class.name}'"
            end
        end

        fill_in_local_headers(response)
        response.body = serialize(response)

        trace_log(:send, :metadata => response.metadata,
                         :payload => response.payload) if trace?
        response
      end

      #
      # converts the received blob into the local representation of an object
      #
      def wire_to_local(body, headers, metadata= {})
        response = LocalResponse.new

        # for tests, we return an OpenStruct if the headers is not a
        # AMQP::Headers
        response.metadata = (headers.is_a?(Hash) ? OpenStruct.new(headers) : headers)
        response.payload = deserialize(body, headers)

        trace_log(:receive, :attributes => (response.metadata.attributes.nil? ?
                                            response.metadata :
                                            response.metadata.attributes),
                            :payload => response.payload,
                            :metadata => metadata) if trace?

        response.payload = build_message(response) if @message

        response
      end

      private
      #######

      def serialize(response)

        mime_type = Tomte::MIME::Type.new(Serializer.content_type_for(@serialization))

        body =
          case response.payload
            when Tomte::Document
            mime_type.add_parameters(:encapsulation => 'tomte-serialized-document')
            Serializer.encode(response.payload, @serialization)
          when Hash
            mime_type.add_parameters(:encapsulation => 'tomte-serialized-object')
            Serializer.encode(response.payload.to_tomte, @serialization)
          else
            mime_type.add_parameters(:encapsulation => 'tomte-serialized-object')
            Serializer.encode(response.payload, @serialization) unless response.payload.nil?
          end
        response.metadata[:content_type] = mime_type.content_type
        body
      end


      def deserialize(body, headers)

        content_type =
          begin
            headers.content_type
          rescue
            headers[:content_type]
          end

        mime_type = Tomte::MIME::Type.new(content_type)
        message_type = mime_type.parameters[:encapsulation]

        payload = Serializer.decode_with_type(body, mime_type.simplified)

        case message_type
          when 'tomte-serialized-document'
            Tomte::Document::Parser.load(payload)
          when 'tomte-serialized-object'
            # XXX it is now all the same?
#            HashWithIndifferentAccess.new(payload)
            Tomte::Document::Parser.load(HashWithIndifferentAccess.new(payload))
          when nil
            payload
        else
          raise TypeError, "Don't know how to handle a message of type #{message_type}"
        end
      end

      def build_message(response)

        m = Tomte::Message.new(:payload => response.payload,
#                           :headers => Tomte::OpenDocument.new(response.metadata.headers),
                           :metadata => Tomte::Message::Metadata.new(response.metadata.attributes))
        m.headers = Tomte::OpenDocument.new(response.metadata.headers)

        tomte_debug do
          log.debug m.ai.subindent
        end

        m
      end
    end
  end
end
