module Tomte::Protocol
  module RPC

    class Server < Tomte::Protocol::Dialog::Server

      # all the methods starting with _ should not go trough
      DEFAULT_METHOD_FILTER = lambda {|method| method.to_s =~ /^_/}

      # method called on message to determine the method
      DEFAULT_KEY_BINDING = ''

      # default options for the api exchange
      DEFAULT_EXCHANGE_OPTIONS = { }.freeze

      # if this list contains a methods list, only these are allowed to be called
      attr_accessor :allowed_methods

      # a list of methods that will be called using a deferred procedure
      attr_accessor :deferred_methods

      # a list of methods that will be called with a raw signature (ie, they only take a request object, instead
      # of a full list of arguments)
      attr_accessor :raw_methods


      def initialize(connection, options = {})

        options[:queue] = options[:api] if(!options.has_key?(:queue) && options[:api])

        super

        options[:exchange].is_a?(Tomte::Protocol::Exchange) ?
          @exchange = options[:exchange] :
          @exchange_name = options[:exchange]

        @exchange_name =    options[:api] if options[:api]
        @exchange_type =    options.fetch(:exchange_type,     :direct)
        @exchange_options = options.fetch(:exchange_options,  DEFAULT_EXCHANGE_OPTIONS)
        @key =              options.fetch(:key,               DEFAULT_KEY_BINDING)
        @deferred_methods = options.fetch(:deferred_methods,  [])

        raise(Error, "RPC Server, an api point, or an exchange is mandatory") unless @exchange || @exchange_name

        @allowed_methods =   options.fetch(:allowed_methods,  [])
        @method_filter ||=   options.fetch(:method_filter,    DEFAULT_METHOD_FILTER)
        @raw ||=             options.fetch(:raw,              false)
        @raw_methods =       options.fetch(:raw_methods,      [])

        @enveloper = Tomte::Protocol::Enveloper::Generic.new

        @return_request =    options.fetch(:return_request,   false)
        @remote_backtrace =  options[:remote_backtrace] || Tomte.config.protocol[:remote_backtrace] || false

        server options[:server] if options[:server]
        on_call &method(:call_handler)
      end

      # we choose an enveloper
      def enveloper=(*_); end

      def server(srv = nil, &block)
        srv = block if block_given?

        @server =
          case srv
          when ::Class
            srv.new
          when ::Module
            (::Class.new { include(srv) }).new
          else
            srv
          end

        server_service_method :_api_initialize
        @server
      end
      alias_method :server=, :server



      def connect!(options = {}, &block)

        super(options) do

          if(@exchange)
            bind(@exchange, :routing_key => @key) { block.call(self) if block_given? }
          else
            @exchange = declare_exchange(@exchange_name,
                                         @exchange_type,
                                         @exchange_options) do |ex|

              bind(@exchange, :routing_key => @key) { block.call(self) if block_given? }
              server_service_method :_api_connected

            end
          end
        end

        self
      end

      #
      # close this server
      #
      def close(reply_code = 200, reply_text = DEFAULT_REPLY_TEXT, &block)
        if(@exchange)
          unbind(@exchange) { super }
        else
          super
        end
      end


      #
      # delete the internal exchange
      #
      def delete_exchange(options = {}, &block)
        check_ready

        @exchange.delete do |delete_ok|
          log.debug "+ [#{@channel.id}] deleted exchange."
          block.call delete_ok if block
        end
      end

      protected
      #########

      #
      # called on the request, just before sending back the response
      # request.response is already set with anything, result or exception
      # that will be sent back
      #
      def postprocess_request(request)
        request.response.payload.human_output = nil if
            (!request.payload.wants_human_output? &&
            request.response.payload.respond_to?(:human_output=))
      end

      private
      #######

      #
      # handles each RPC call
      #
      def call_handler(request)

        response = Response.new(:request => (@return_request ? request.payload : nil))
        request.response.payload = response

        if request.response_callback
          original_callback = request.response_callback
          request.response_callback = proc do |rq, res|
            if res.is_a? Exception
              rq.response.payload = res
            else
              rq.response.payload.response = res
            end
            original_callback.call(rq, rq.response.payload)
          end
        end

        tomte_debug do
          log.debug "+ [#{@channel.id}] received RPC request:"
          log.debug request.payload.ai.subindent
        end

        verb = sanitize_rpc_request(request.payload)

        if @deferred_methods.include?(verb)
          call_asynchronously(verb, request)
        else
          response.response = call_synchronously(verb, request)
        end
        return response

      rescue Exception => exception
        log.exception exception, :warn

        response.exception = filter_exception(exception)
        #request.response.payload = response
        response = exception
        return response

      # not good, it skews async processing
      #ensure
      #  return response
      end


      def call_asynchronously(verb, request)

        callback = proc { |r| request.return(r) }
        errback = proc { |r| request.error(r) }
        job = proc { call_synchronously(verb, request) }

        protected_defer(job, callback, errback)
        throw :_async
      end

      def call_synchronously(verb, request)

        # FIXME json/bson do not transport symbols as keys
        # but this is an ugly solution
        parameters = request.payload.parameters.map do |p|
          p = p.is_a?(Hash) ? HashWithIndifferentAccess.new(p) : p
        end

        server_service_method :_api_before_request, request

        response =
          if @server.is_a? Proc
            if @raw
              @server.call(verb, request)
            else
              @server.call(verb, parameters)
            end

          else

            if @raw || raw_methods.include?(verb)
              @server.__send__(verb, request)
            else
              @server.__send__(verb, *parameters)
            end
          end

        server_service_method :_api_after_request, request

        response
      end

      def server_service_method(method, *args)

        @server.__send__(method, self, *args) if @server.respond_to?(method)
      #rescue NoMethodError
      end


      def filter_exception(e)
        e.set_backtrace([]) if !@remote_backtrace
        e
      end

      def sanitize_rpc_request(request)
        raise ArgumentError, "Message malformed, missing verb"  unless request.respond_to?(:verb)
        verb = request.verb.to_sym
        unless allowed_methods.blank?
          raise NoMethodError, "No method '#{verb}' for API '#{@server.class.name}'" unless allowed_methods.include?(verb)
        end
        raise NoMethodError, "Filtered verb #{verb} for #{@server.class.name}" if (@method_filter && @method_filter.call(verb))
        verb
      end

    end
  end
end
