# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol
  #
  # performs objects serialization before sending them on a wire
  #
  # supported formats:
  #
  #  :text => ['text/plain'],
  #  :json => ['application/json'],
  #  :bson => ['application/bson', 'application/x-bson']
  #
  module Serializer


    module BSON
      def self.encode(payload, options = {})
        payload.respond_to?(:to_bson) ?
        payload.to_bson :
        ::BSON.serialize(payload, options[:check_key],options[:move_id])
      end

      def self.decode(body)
        ::BSON.deserialize(body)
      end
    end

    module JSON
      def self.encode(payload, options = {})
        ActiveSupport::JSON.encode(payload, options)
      end

      def self.decode(body)
        ActiveSupport::JSON.decode(body)
      end
    end

    module Null
      def self.encode(payload)
        payload.to_s
      end

      def self.decode(body)
        body
      end
    end

    SERIALIZERS = { :json => JSON, :bson => BSON, :null => Null, :text => Null }.freeze
    SERIALIZERS_CONTENT_TYPES = {
        :text => ['text/plain'],
        :json => ['application/json', 'application/x-json'],
        :bson => ['application/x-tomte-bson', 'application/tomte-bson',
                  'application/x-bson', 'application/bson'],
        :null => [ 'application/octet-stream']
    }.freeze

    class Error < StandardError; end
    class FormatError < Error; end

    class SerializationError < Error
      attr_accessor :action, :payload, :type, :rescued_exception
      def initialize(action, payload, type = nil, rescued_exception = nil)
        @action, @packet, @type, @caught_exception = action, payload, type, rescued_exception
        super("Could not #{action} payload, requested type: '#{type}'\n #{rescued_exception}")
      end
    end

    def self.serializer_for(format)
      (SERIALIZERS[format.to_sym] || raise(FormatError, "Unknown Serializer for format '#{format}'"))
    end

    def self.encode(payload, format = :json)
      serializer_for(format).encode(payload)
    rescue Exception => e      
      Tomte.logger.try(:error, e)
      Tomte.logger.try(:error, "\t" + e.backtrace.join("\n\t"))
      raise SerializationError.new(:encode, payload, format, e)
    end

    def self.decode(body, format)
      serializer_for(format).decode(body)
    rescue Exception => e 
      Tomte.logger.try(:error, e)
      Tomte.logger.try(:error, "\t" + e.backtrace.join("\n\t"))
      raise SerializationError.new(:decode, body, format, e)
    end

    def self.decode_with_type(body, content_type)
      decode(body, format_for(content_type))

    end

    def self.content_type_for(format)
      (SERIALIZERS_CONTENT_TYPES[format.to_sym] ||
          raise(FormatError, "Unknown Serialization Format '#{format}'")).first
    end

    def self.format_for(content_type)
      SERIALIZERS_CONTENT_TYPES.each do |name, types|
         return name if types.include?(content_type)
      end
      raise FormatError, "Content type '#{content_type}' is not known to the serializer"
    end

  end
end
